/**
 * Copyright (c) 2010-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the New BSD License (3-clause license).
 *
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the New BSD License (3-clause license)
 * for more details.
 *
 * You should have received a copy of the New BSD License (3-clause license)
 * along with this program/library; If not, see http://directory.fsf.org/wiki/License:BSD_3Clause/
 * for the New BSD License (3-clause license).
 */
package com.ebmwebsourcing.easycommons.io;

import static org.junit.Assert.assertArrayEquals;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.channels.Channel;
import java.nio.channels.Channels;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.FileChannel;
import java.nio.channels.ReadableByteChannel;
import java.nio.channels.WritableByteChannel;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.TemporaryFolder;

/**
 * @author Marc Jambert - EBM WebSourcing
 */
public class ByteChannelDumpHelperTest {

    private static final int[] TEST_INPUT_DATA_SIZES = new int[] { 1, 2, 3, 8, 16, 32, 47 };

    private static final int[] TEST_BYTE_BUFFER_SIZES = new int[] { 1, 2, 4, 64, 65 };

    @Rule
    public final TemporaryFolder temporaryFolder = new TemporaryFolder();

    @Test
    public void testCreateDumpFileChannelWithNonEmptyExistingFile() throws Exception {
        File dumpFile = this.temporaryFolder.newFile("dump-file");
        FileOutputStream fos = new FileOutputStream(dumpFile);
        fos.write(1);
        fos.close();
        FileChannel fileChannel = ByteChannelDumpHelper.createDumpFileChannel(dumpFile);
        checkFileExistsAndEmpty(dumpFile);
        checkFileChannelIsReadyAndEmpty(fileChannel);
    }

    @Test
    public void testCreateDumpFileChannelWithNonExistingDumpFile() throws Exception {
        File dumpFile = createNonExistingTempFile();
        FileChannel fileChannel = ByteChannelDumpHelper.createDumpFileChannel(dumpFile);
        checkFileExistsAndEmpty(dumpFile);
        checkFileChannelIsReadyAndEmpty(fileChannel);
    }

    @Test
    public void testCreateDumpFileChannelWithNonExistingDumpFileDirectory() throws Exception {
        File dumpFile = createTempFileInANonExistingDirectory();
        FileChannel fileChannel = ByteChannelDumpHelper.createDumpFileChannel(dumpFile);
        checkFileExistsAndEmpty(dumpFile);
        checkFileChannelIsReadyAndEmpty(fileChannel);
    }

    private final File createNonExistingTempFile() throws IOException {

        final File nonExistingTempFile = this.temporaryFolder.newFile("non-existing-temp-file");
        if (!nonExistingTempFile.delete()) {
            throw new IOException(String.format("Unable to delete the file '%s'.",
                    nonExistingTempFile.getAbsolutePath()));
        } else {
            return nonExistingTempFile;
        }
    }

    private final File createTempFileInANonExistingDirectory() throws IOException {

        final File tempFileInNonExistingDir = this.temporaryFolder.newFile("temp-file-in-non-existing-directory");
        FileSystemHelper.forceDelete(tempFileInNonExistingDir);
        return tempFileInNonExistingDir;
    }

    private void checkFileChannelIsReadyAndEmpty(FileChannel fileChannel) throws IOException {
        assertTrue(fileChannel.isOpen());
        assertEquals(0, fileChannel.size());
        assertEquals(0, fileChannel.position());
    }

    private void checkFileExistsAndEmpty(File file) {
        assertTrue(file.exists());
        assertTrue(file.length() == 0);
    }

    @Test
    public void testReadBufferByBufferFromVariousInputsWithVariousBufferSizes()
            throws Exception {
        testProcessBufferByBufferFromVariousInputsWithVariousBufferSizes(new BufferByBufferReadTesterBuilder());
    }

    @Test
    public void testWriteBufferByBufferFromVariousInputsWithVariousBufferSizes()
            throws Exception {
        testProcessBufferByBufferFromVariousInputsWithVariousBufferSizes(new BufferByBufferWriteTesterBuilder());
    }

    private void testProcessBufferByBufferFromVariousInputsWithVariousBufferSizes(
            AbstractMethodTesterBuilder testerBuilder) throws Exception {
        for (int i : TEST_INPUT_DATA_SIZES) {
            testerBuilder.setTestInputData(generateTestInputData(i));
            for (int j : TEST_BYTE_BUFFER_SIZES) {
                testerBuilder.setByteBuffer(new byte[j]);
                testerBuilder.setDumpFile(this.temporaryFolder.newFile("dump-file" + i + "-" + j));
                for (int testOffset : generateTestOffsets(j)) {
                    testerBuilder.setOffset(testOffset);
                    for (int testLen : generateTestLens(j, testOffset)) {
                        testerBuilder.setLen(testLen);
                        AbstractMethodTester tester = testerBuilder.build();
                        tester.testCallUntilEndOfStream();
                    }
                }
            }
        }
    }

    @Test
    public void testReadBufferByBufferFromEmptyInput() throws Exception {
        BufferByBufferReadTesterBuilder testerBuilder = new BufferByBufferReadTesterBuilder();
        testerBuilder.setByteBuffer(new byte[4]);
        testerBuilder.setOffset(1);
        testerBuilder.setLen(1);
        testReadFromEmptyInput(testerBuilder);
    }

    private void testReadFromEmptyInput(AbstractMethodTesterBuilder testerBuilder)
            throws Exception {
        testerBuilder.setDumpFile(this.temporaryFolder.newFile("dump-file"));
        testerBuilder.setTestInputData(new byte[0]);
        AbstractMethodTester tester = testerBuilder.build();
        assertEquals(-1, tester.callMethodOnce());
    }

    @Test(expected = ClosedChannelException.class)
    public void testReadBufferByBufferFromClosedInput() throws Exception {
        BufferByBufferReadTesterBuilder testerBuilder = new BufferByBufferReadTesterBuilder();

        testProcessOnClosedStream(testerBuilder);
    }

    @Test(expected = ClosedChannelException.class)
    public void testWriteBufferByBufferOnClosedOutput() throws Exception {
        BufferByBufferWriteTesterBuilder testerBuilder = new BufferByBufferWriteTesterBuilder();
      
        testProcessOnClosedStream(testerBuilder);
    }

    private void testProcessOnClosedStream(AbstractMethodTesterBuilder testerBuilder)
            throws Exception {
        testerBuilder.setByteBuffer(new byte[1]);
        testerBuilder.setOffset(0);
        testerBuilder.setLen(1);
        testerBuilder.setDumpFile(this.temporaryFolder.newFile("dump-file"));
        testerBuilder.setTestInputData(new byte[] { 0x01, 0x02 });
    
        AbstractMethodTester tester = testerBuilder.build();
        tester.getChannelToDump().close();
        tester.callMethodOnce();
    }

    private byte[] generateTestInputData(int length) {
        byte[] testInputData = new byte[length];
        for (int i = 0; i < testInputData.length; ++i) {
            testInputData[i] = (byte) i;
        }
        return testInputData;
    }

    private final Integer[] generateTestLens(int j, int testOffset) {
        Set<Integer> testLens = new HashSet<Integer>();
        for (int testLen : new int[] { j - testOffset, (j - testOffset) / 2, 1 }) {
            if (testLen == 0)
                continue;
            testLens.add(testLen);
        }
        return testLens.toArray(new Integer[testLens.size()]);
    }

    private final Integer[] generateTestOffsets(int j) {
        Set<Integer> testOffsets = new HashSet<Integer>();
        for (int testOffset : new int[] { 0, j / 2, j - 1 }) {
            if (testOffset >= j - 1)
                continue;
            testOffsets.add(testOffset);
        }
        return testOffsets.toArray(new Integer[testOffsets.size()]);
    }

    private static class BufferByBufferReadTesterBuilder extends
            AbstractMethodTesterBuilder {

        @Override
        protected AbstractMethodTester build() throws IOException {
            return new BufferByBufferReadTester(byteBuffer, offset, len, dumpFile,
                    testInputData);
        }

    }

    private static class BufferByBufferWriteTesterBuilder extends
    AbstractMethodTesterBuilder {
    
        @Override
        protected AbstractMethodTester build() throws IOException {
            return new BufferByBufferWriteTester(byteBuffer, offset, len, dumpFile,
                    testInputData);
        }
    
    }

    private static abstract class AbstractMethodTesterBuilder {
    
        protected byte[] testInputData;
    
        protected File dumpFile;
        
        protected byte[] byteBuffer;
        
        protected int offset;
    
        protected int len;    

    
        public AbstractMethodTesterBuilder() {
            testInputData = null;
            dumpFile = null;
            
            this.byteBuffer = null;
            this.offset = 0;
            this.len = 0;
        }
    
        public final void setTestInputData(byte[] testInputData) {
            this.testInputData = testInputData;
        }
    
        public final void setDumpFile(File dumpFile) {
            this.dumpFile = dumpFile;
        }
        
        public final void setOffset(int offset) {
            this.offset = offset;
        }
    
        public final void setLen(int len) {
            this.len = len;
        }
    
        public final void setByteBuffer(byte[] byteBuffer) {
            this.byteBuffer = byteBuffer;
        }
    
        protected abstract AbstractMethodTester build() throws IOException;
    }

    private static class BufferByBufferReadTester extends
            AbstractBufferByBufferMethodTester {

        protected byte[] byteBufferBeforeReadCall;

        public BufferByBufferReadTester(byte[] byteBuffer, int offset, int len,
                File dumpFile, byte[] testInputData) throws IOException {
            super(byteBuffer, offset, len, dumpFile, testInputData);
            this.byteBufferBeforeReadCall = null;
        }

        @Override
        protected ReadableByteChannel createTestChannelToDump() {
            return Channels.newChannel(new ByteArrayInputStream(getTestInputData()));
        }

        @Override
        protected void updateSoFarProcessedBytes(byte[] lastProcessedBytes) throws IOException {
            getSoFarProcessedBytesOutputStream().write(lastProcessedBytes);
        }

        @Override
        public int callMethodOnce() throws IOException {
            byteBufferBeforeReadCall = Arrays.copyOf(getByteBuffer(), getByteBuffer().length);

            ByteBuffer testBuffer = ByteBuffer.wrap(getByteBuffer(), getOffset(), getLen());

            int lastNbRead = ByteChannelDumpHelper.dumpAndRead(
                    (ReadableByteChannel) getChannelToDump(), testBuffer, getDumpChannel());

            if (lastNbRead != -1) {
                setLastProcessedBytes(Arrays.copyOfRange(getByteBuffer(), getOffset(), getOffset()
                        + lastNbRead));
            }
            return lastNbRead;
        }

        @Override
        protected void check() throws IOException {
            super.check();
            checkByteBufferUntouchedBeforeDestBytes();
            checkByteBufferUntouchedAfterDestBytes();
        }

        private void checkByteBufferUntouchedBeforeDestBytes() {
            byte[] originalBytesBeforeDestBytes = Arrays.copyOfRange(byteBufferBeforeReadCall, 0,
                    getOffset());
            byte[] modifiedBytesBeforeDestBytes = Arrays.copyOfRange(getByteBuffer(), 0,
                    getOffset());
            assertArrayEquals(originalBytesBeforeDestBytes, modifiedBytesBeforeDestBytes);
        }

        private void checkByteBufferUntouchedAfterDestBytes() {
            byte[] originalBytesAfterDestBytes = Arrays.copyOfRange(byteBufferBeforeReadCall,
                    getOffset() + getLen(), byteBufferBeforeReadCall.length);
            byte[] modifiedBytesAfterDestBytes = Arrays.copyOfRange(getByteBuffer(), getOffset()
                    + getLen(), byteBufferBeforeReadCall.length);
            assertArrayEquals(originalBytesAfterDestBytes, modifiedBytesAfterDestBytes);
        }

    }

    private static class BufferByBufferWriteTester extends
            AbstractBufferByBufferMethodTester {
    
        public BufferByBufferWriteTester(byte[] byteBuffer, int offset, int len,
                File dumpFile, byte[] testInputData) throws IOException {
            super(byteBuffer, offset, len, dumpFile, testInputData);
        }
    
        @Override
        protected void updateSoFarProcessedBytes(byte[] lastProcessedBytes) throws IOException {
        }
    
        @Override
        protected WritableByteChannel createTestChannelToDump() {
            return Channels.newChannel(getSoFarProcessedBytesOutputStream());
        }
    
        protected int doCallMethodOnceIfBufferCanBeFullyFilled() throws IOException {
            
            System.arraycopy(getTestInputData(), getSoFarProcessedBytesOutputStream().size(),
                    getByteBuffer(), getOffset(), getLen());
            
            ByteChannelDumpHelper.dumpAndWrite((WritableByteChannel) getChannelToDump(), 
                    ByteBuffer.wrap(getByteBuffer(), getOffset(), getLen()), getDumpChannel());
    
            setLastProcessedBytes(Arrays.copyOfRange(getByteBuffer(), getOffset(), getOffset()
                    + getLen()));
            return getLen();
        }
    
        @Override
        public final int callMethodOnce() throws IOException {
            int nbSoFarWrittenBytes = getSoFarProcessedBytesOutputStream().size();
            if (nbSoFarWrittenBytes == getTestInputData().length) {
                return -1;
            }
            int remainingNbBytesToWrite = getTestInputData().length - nbSoFarWrittenBytes;
            if ((getOffset() + getLen() <= getByteBuffer().length)
                    && (remainingNbBytesToWrite < getLen())) {
                // test is no more significant, stop one chunk before.
                return -1;
            }
            return doCallMethodOnceIfBufferCanBeFullyFilled();
        }
    
    }

    private static abstract class AbstractBufferByBufferMethodTester extends
            AbstractMethodTester {
    
        private final byte[] byteBuffer;
    
        private final int offset;
    
        private final int len;
    
        protected AbstractBufferByBufferMethodTester(byte[] byteBuffer, int offset,
                int len, File dumpFile, byte[] testInputData) throws IOException {
            super(dumpFile, testInputData);
            assert byteBuffer != null;
            this.byteBuffer = byteBuffer;
            this.offset = offset;
            this.len = len;
        }
    
        protected final byte[] getByteBuffer() {
            return byteBuffer;
        }
    
        protected final int getOffset() {
            return offset;
        }
    
        protected final int getLen() {
            return len;
        }
    }

    private static abstract class AbstractMethodTester {
    
        private final byte[] testInputData;
    
        private final File dumpFile;
    
        private Channel channelToDump;
    
        private final FileChannel dumpChannel;
    
        private final ByteArrayOutputStream soFarProcessedBytesOutputStream;
    
        private byte[] lastProcessedBytes;
    
        protected AbstractMethodTester(File dumpFile, byte[] testInputData) throws IOException {
            super();
            assert dumpFile != null;
            assert testInputData != null;
            this.testInputData = testInputData;
            this.dumpFile = dumpFile;
            this.channelToDump = null;
            this.dumpChannel = ByteChannelDumpHelper.createDumpFileChannel(dumpFile);
            this.soFarProcessedBytesOutputStream = new ByteArrayOutputStream();
            this.lastProcessedBytes = null;
        }
    
        protected final void setLastProcessedBytes(byte[] lastProcessedBytes) {
            this.lastProcessedBytes = lastProcessedBytes;
        }
    
        protected abstract void updateSoFarProcessedBytes(byte[] lastProcessedBytes)
                throws IOException;
    
        protected final ByteArrayOutputStream getSoFarProcessedBytesOutputStream() {
            return soFarProcessedBytesOutputStream;
        }
    
        protected final byte[] getTestInputData() {
            return testInputData;
        }
    
        public abstract int callMethodOnce() throws IOException;
    
        private final void checkProcessedAndDumpedBytes() throws IOException {
            byte[] soFarWrittenBytes = this.soFarProcessedBytesOutputStream.toByteArray();
            byte[] dumpedBytes = IOHelper.readFileAsBytes(dumpFile);
            assertArrayEquals(Arrays.copyOfRange(testInputData, 0, soFarWrittenBytes.length),
                    soFarWrittenBytes);
            assertArrayEquals(soFarWrittenBytes, dumpedBytes);
        }
    
        protected final Channel getChannelToDump() {
            if (channelToDump == null) {
                channelToDump = createTestChannelToDump();
            }
            return channelToDump;
        }
    
        protected final FileChannel getDumpChannel() {
            return dumpChannel;
        }
    
        protected abstract Channel createTestChannelToDump();
    
        protected void check() throws IOException {
            checkProcessedAndDumpedBytes();
        }
    
        public final void testCallUntilEndOfStream() throws IOException {
            while (true) {
                setLastProcessedBytes(null);
                callMethodOnce();
                if (lastProcessedBytes == null)
                    break;
                updateSoFarProcessedBytes(lastProcessedBytes);
                check();
            }
        }
    
    }

}
