/**
 * Copyright (c) 2010-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the New BSD License (3-clause license).
 *
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the New BSD License (3-clause license)
 * for more details.
 *
 * You should have received a copy of the New BSD License (3-clause license)
 * along with this program/library; If not, see http://directory.fsf.org/wiki/License:BSD_3Clause/
 * for the New BSD License (3-clause license).
 */
package com.ebmwebsourcing.easycommons.io;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import java.io.File;
import java.io.IOException;

import org.apache.commons.lang3.SystemUtils;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;
import org.junit.rules.TemporaryFolder;

import com.ebmwebsourcing.easycommons.lang.reflect.ReflectionHelper;

/**
 * @author Marc Jambert - EBM WebSourcing
 * @author Christophe DENEUX - Linagora
 */
public class FileSystemHelperTest {

    private static final String FOLDER_PATH_LONG = File.separator + "root" + File.separator + "dir1" + File.separator
            + "dir2" + File.separator + "dir3" + File.separator;

    private static final String FOLDER_PATH_SHORT = File.separator + "root" + File.separator + "dir1" + File.separator;

    private static final String FILE_PATH_SHORT = File.separator + "root" + File.separator + "dir1" + File.separator
            + "pom.xml";

    private static final String FILE_PATH_LONG = File.separator + "root" + File.separator + "dir1" + File.separator
            + "dir2" + File.separator + "dir3" + File.separator + "pom.xml";

    private static final String NOT_RELATIVE_FILE_PATH = File.separator + "notrelative" + File.separator + "dir1"
            + File.separator + "pom.xml";

    private static final String NOT_RELATIVE_FOLDER_PATH = File.separator + "notrelative" + File.separator
            + "dir1";

    @Rule
    public TemporaryFolder tempFolder = new TemporaryFolder();

    @Rule
    public ExpectedException expectedException = ExpectedException.none();

    @Test(expected = IllegalArgumentException.class)
    public void testGetRelativePathIllegalArgumentException_Deprecated() {
        FileSystemHelper.getRelativePath("/root/dir1", FOLDER_PATH_SHORT);
    }

    @Test
    public void testGetRelativePathFolderToFolder_Deprecated() {
        assertEquals(".." + File.separator + ".." + File.separator,
                FileSystemHelper.getRelativePath(FOLDER_PATH_LONG, FOLDER_PATH_SHORT));
    }

    @Test
    public void testGetRelativePathFolderToFile_Deprecated() {
        assertEquals(".." + File.separator + ".." + File.separator + "pom.xml",
                FileSystemHelper.getRelativePath(FOLDER_PATH_LONG, FILE_PATH_SHORT));
    }

    @Test
    public void testGetRelativePathFolderToFolderForward_Deprecated() {
        assertEquals("dir2" + File.separator + "dir3" + File.separator,
                FileSystemHelper.getRelativePath(FOLDER_PATH_SHORT, FOLDER_PATH_LONG));
    }

    @Test
    public void testGetRelativePathFolderToFileForward_Deprecated() {
        assertEquals("dir2" + File.separator + "dir3" + File.separator + "pom.xml",
                FileSystemHelper.getRelativePath(FOLDER_PATH_SHORT, FILE_PATH_LONG));
    }

    @Test
    public void testGetRelativePath_SourceIsNull() throws Exception {

        final String errorMsg = (String) ReflectionHelper.getPrivateFieldValue(
                FileSystemHelper.class, null, "ILLEGAL_ARGUMENT_EXCEPTION_MSG_SRC_NULL");
        this.expectedException.expect(IllegalArgumentException.class);
        this.expectedException.expectMessage(errorMsg);

        final File targetFile = this.tempFolder.newFolder("target");
        FileSystemHelper.getRelativePath(null, targetFile);
    }

    @Test
    public void testGetRelativePath_SourceIsNotAbsolute() throws Exception {

        final String errorMsg = (String) ReflectionHelper.getPrivateFieldValue(
                FileSystemHelper.class, null, "ILLEGAL_ARGUMENT_EXCEPTION_MSG_SRC_NOT_ABSOLUTE");
        this.expectedException.expect(IllegalArgumentException.class);
        this.expectedException.expectMessage(errorMsg);

        final File targetFile = this.tempFolder.newFolder("target");
        FileSystemHelper.getRelativePath(new File("root/dir1"), targetFile);
    }

    @Test
    public void testGetRelativePath_SourceIsNotDirectory() throws IOException {

        final String errorMsg = (String) ReflectionHelper.getPrivateFieldValue(
                FileSystemHelper.class, null, "ILLEGAL_ARGUMENT_EXCEPTION_MSG_SRC_NOT_DIR");
        this.expectedException.expect(IllegalArgumentException.class);
        this.expectedException.expectMessage(errorMsg);

        final File srcDir = this.tempFolder.newFile("source.file");
        FileSystemHelper.getRelativePath(srcDir, new File("/root/dir1"));
    }

    @Test
    public void testGetRelativePath_TargetIsNull() throws IOException {

        final String errorMsg = (String) ReflectionHelper.getPrivateFieldValue(
                FileSystemHelper.class, null, "ILLEGAL_ARGUMENT_EXCEPTION_MSG_TARGET_NULL");
        this.expectedException.expect(IllegalArgumentException.class);
        this.expectedException.expectMessage(errorMsg);

        final File srcDir = this.tempFolder.newFile("source.file");
        FileSystemHelper.getRelativePath(srcDir, null);
    }

    @Test
    public void testGetRelativePath_TargetIsNotAbsolute() throws Exception {

        final String errorMsg = (String) ReflectionHelper.getPrivateFieldValue(
                FileSystemHelper.class, null, "ILLEGAL_ARGUMENT_EXCEPTION_MSG_TARGET_NOT_ABSOLUTE");
        this.expectedException.expect(IllegalArgumentException.class);
        this.expectedException.expectMessage(errorMsg);

        final File srcDir = this.tempFolder.newFolder("source");
        FileSystemHelper.getRelativePath(srcDir, new File("target"));
    }

    /**
     * <p>
     * Check {@link FileSystemHelper#getRelativePath(File, File)} to get the
     * relative path of <code>/root/dir1</code> against
     * <code>/root/dir1/dir2/dir3</code>
     * </p>
     * <p>
     * Expected result: <code>../..</code>
     */
    @Test
    public void testGetRelativePathFolderToFolder() throws Exception {
        final File root = this.tempFolder.newFolder("root");
        final File dir1 = new File(root, "dir1");
        final File dir2 = new File(dir1, "dir2");
        final File dir3 = new File(dir2, "dir3");
        assertTrue("Unable to create directory tree", dir3.mkdirs());
        
        assertEquals("Unexpected relative path", ".." + File.separator + ".." + File.separator,
                FileSystemHelper.getRelativePath(dir3, dir1));
    }

    @Test
    public void testGetRelativePathFolderToFile() throws Exception {
        final File root = this.tempFolder.newFolder("root");
        final File dir1 = new File(root, "dir1");
        final File file = new File(dir1, "pom.xml");
        final File dir2 = new File(dir1, "dir2");
        final File dir3 = new File(dir2, "dir3");
        assertTrue("Unable to create directory tree", dir3.mkdirs());

        assertEquals("Unexpected relative path", ".." + File.separator + ".." + File.separator + "pom.xml",
                FileSystemHelper.getRelativePath(dir3, file));
    }

    @Test
    public void testGetRelativePathFolderToFolderForward() throws Exception {
        final File root = this.tempFolder.newFolder("root");
        final File dir1 = new File(root, "dir1");
        final File dir2 = new File(dir1, "dir2");
        final File dir3 = new File(dir2, "dir3");
        assertTrue("Unable to create directory tree", dir3.mkdirs());

        assertEquals("Unexpected relative path", "." + File.separator + "dir2" + File.separator + "dir3",
                FileSystemHelper.getRelativePath(dir1, dir3));
    }

    @Test
    public void testGetRelativePathFolderToFileForward() throws Exception {
        final File root = this.tempFolder.newFolder("root");
        final File dir1 = new File(root, "dir1");
        final File dir2 = new File(dir1, "dir2");
        final File dir3 = new File(dir2, "dir3");
        final File file = new File(dir3, "pom.xml");
        assertTrue("Unable to create directory tree", dir3.mkdirs());

        assertEquals("Unexpected relative path", "." + File.separator + "dir2" + File.separator + "dir3"
                + File.separator + "pom.xml",
                FileSystemHelper.getRelativePath(dir1, file));
    }

    @Test
    public void testGetRelativePathNotRelativeFile() throws Exception {
        final File root = this.tempFolder.newFolder("root");
        final File dir1 = new File(root, "dir1");
        assertTrue("Unable to create directory tree", dir1.mkdirs());

        assertEquals(
                (SystemUtils.IS_OS_WINDOWS ? "X:" : "") + NOT_RELATIVE_FILE_PATH,
                FileSystemHelper.getRelativePath(dir1, new File((SystemUtils.IS_OS_WINDOWS ? "X:" : "")
                        + NOT_RELATIVE_FILE_PATH)));
    }

    @Test
    public void testGetRelativePathNotRelativeFolder() throws Exception {
        final File root = this.tempFolder.newFolder("root");
        final File dir1 = new File(root, "dir1");
        assertTrue("Unable to create directory tree", dir1.mkdirs());

        assertEquals(
                (SystemUtils.IS_OS_WINDOWS ? "X:" : "") + NOT_RELATIVE_FOLDER_PATH,
                FileSystemHelper.getRelativePath(dir1, new File((SystemUtils.IS_OS_WINDOWS ? "X:" : "")
                        + NOT_RELATIVE_FOLDER_PATH)));
    }

    @Test
    public void testCreateTempDirReallyCreateADir() throws Exception {
        File tempDir = FileSystemHelper.createTempDir();
        assertTrue(tempDir.isDirectory());
    }

    @Test
    public void testCreateTempDirTwiceReturnDifferentTempDir() throws Exception {
        File tempDir = FileSystemHelper.createTempDir();
        File tempDir2 = FileSystemHelper.createTempDir();
        assertFalse(tempDir.equals(tempDir2));
    }

    @Test
    public void testCreateTempDirUsesSystemTempDirByDefault() throws Exception {
        File tempDir = FileSystemHelper.createTempDir();
        assertTrue(tempDir.getAbsolutePath().startsWith(System.getProperty("java.io.tmpdir")));
    }

    @Test(expected = IOException.class)
    public void testCreateTempDirInParentDirWhichDoesNotExist() throws Exception {
        File parentDir = new File("/this/certainly/does/not/exist");
        FileSystemHelper.createTempDir(parentDir, "dummyPrefix");
    }

    @Test
    public void testCreateTempDirWithPrefix() throws Exception {
        String testPrefix = "testPrefix";
        File tempDir = FileSystemHelper.createTempDir(testPrefix);
        assertTrue(tempDir.getName().startsWith(testPrefix));
    }

    @Test
    public void testCreateTempCreateAnEmptyDir() throws Exception {
        File tempDir = FileSystemHelper.createTempDir();
        assertTrue(tempDir.list().length == 0);
    }

    @Test(expected = IOException.class)
    public void testCleanNotExistingDirectory() throws IOException {
        File notExistingDirectory = createNotExistingDirectory();
        FileSystemHelper.cleanDirectory(notExistingDirectory);
    }

    @Test(expected = IOException.class)
    public void testCleanDirectoryThatIsntADirectory() throws IOException {
        File notADirectory = new File(FileSystemHelper.createTempDir(), "notADirectory.bin");
        FileSystemHelper.cleanDirectory(notADirectory);
    }

    private File createNotExistingDirectory() throws IOException {
        File notExistingDirectory = FileSystemHelper.createTempDir();
        notExistingDirectory.delete();
        assertFalse(notExistingDirectory.exists());
        return notExistingDirectory;
    }

    @Test
    public void testCleanEmptyDirectoryDontModifyIt() throws IOException {
        File emptyDirectory = FileSystemHelper.createTempDir();

        testDirectoryCleaned(emptyDirectory);
    }

    @Test
    public void testCleanDirectoryContainingNotDirectoryFiles() throws IOException {
        File directoryFullOfFiles = createDirectoryFullOfFiles();

        testDirectoryCleaned(directoryFullOfFiles);
    }

    private File createDirectoryFullOfFiles() throws IOException {
        File directoryFullOfFiles = FileSystemHelper.createTempDir();
        File subFile1 = new File(directoryFullOfFiles, "subFile1.txt");
        subFile1.createNewFile();
        File subFile2 = new File(directoryFullOfFiles, "subFile2.txt");
        subFile2.createNewFile();
        return directoryFullOfFiles;
    }

    private void testDirectoryCleaned(File directory) throws IOException {
        FileSystemHelper.cleanDirectory(directory);

        assertTrue(directory.exists());
        assertTrue(directory.list().length == 0);
    }

    @Test
    public void testCleanDirectoryContainingNonEmptyDirectories() throws IOException {
        File directoryFullOfDirectories = createDirectoryFullOfNotEmptyDirectories();

        testDirectoryCleaned(directoryFullOfDirectories);
    }

    @Test
    public void testCleanDirectoryContainingNonEmptyDirectoriesAndFiles() throws IOException {
        File directoryFullOfDirectories = createDirectoryFullOfNotEmptyDirectoriesAndFiles();

        testDirectoryCleaned(directoryFullOfDirectories);
    }

    private File createDirectoryFullOfNotEmptyDirectories() throws IOException {
        File directoryFullOfDirectories = FileSystemHelper.createTempDir();
        File subDir1 = new File(directoryFullOfDirectories, "subDir");
        subDir1.mkdir();
        File subDir2 = new File(directoryFullOfDirectories, "subDir2");
        subDir2.mkdir();
        File subFile1 = new File(subDir1, "subFile1.txt");
        subFile1.createNewFile();
        File subFile2 = new File(subDir2, "subFile2.txt");
        subFile2.createNewFile();
        return directoryFullOfDirectories;
    }

    private File createDirectoryFullOfNotEmptyDirectoriesAndFiles() throws IOException {
        File directoryFullOfDirectories = createDirectoryFullOfNotEmptyDirectories();
        File subFile1 = new File(directoryFullOfDirectories, "subFile1.txt");
        subFile1.createNewFile();
        File subFile2 = new File(directoryFullOfDirectories, "subFile2.txt");
        subFile2.createNewFile();
        return directoryFullOfDirectories;
    }

    @Test
    public void testForceDeleteNotExistingFile() throws IOException {
        File notExistingDirectory = createNotExistingDirectory();
        FileSystemHelper.forceDelete(notExistingDirectory);
        assertFalse(notExistingDirectory.exists());
    }

    @Test
    public void testForceDeleteEmptyDirectory() throws IOException {
        File emptyDirectory = FileSystemHelper.createTempDir();

        testFileDeleted(emptyDirectory);
    }

    @Test
    public void testForceDeleteDirectoryContainingNotDirectoryFiles() throws IOException {
        File directoryFullOfFiles = createDirectoryFullOfFiles();

        testFileDeleted(directoryFullOfFiles);
    }

    @Test
    public void testForceDeleteDirectoryContainingNotEmptyDirectories() throws IOException {
        File directoryFullOfNotEmptyDirectories = createDirectoryFullOfNotEmptyDirectories();

        testFileDeleted(directoryFullOfNotEmptyDirectories);
    }

    @Test
    public void testForceDeleteDirectoryContainingNotEmptyDirectoriesAndFiles() throws IOException {
        File directoryFullOfNotEmptyDirectories = createDirectoryFullOfNotEmptyDirectoriesAndFiles();

        testFileDeleted(directoryFullOfNotEmptyDirectories);
    }

    private void testFileDeleted(File file) throws IOException {
        FileSystemHelper.forceDelete(file);

        assertFalse(file.exists());
    }
}
