/**
 * Copyright (c) 2010-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the New BSD License (3-clause license).
 *
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the New BSD License (3-clause license)
 * for more details.
 *
 * You should have received a copy of the New BSD License (3-clause license)
 * along with this program/library; If not, see http://directory.fsf.org/wiki/License:BSD_3Clause/
 * for the New BSD License (3-clause license).
 */
package com.ebmwebsourcing.easycommons.io;

import static org.junit.Assert.assertArrayEquals;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.io.File;
import java.io.IOException;
import java.net.URL;

import org.junit.Test;

/**
 * @author Marc Jambert - EBM WebSourcing
 */
public class IOHelperTest {

	@Test
	public void testGetLine() throws Exception {
		URL url = Thread.currentThread().getContextClassLoader()
				.getResource("fileToRead.txt");
		File file = new File(url.toURI());
        String line = IOHelper.getLine(file, 18);
        assertEquals("---line18---", line);
	}

	@Test
	public void testGetLineWithAWrongLineNumber() throws Exception {
		URL url = Thread.currentThread().getContextClassLoader()
				.getResource("fileToRead.txt");
		File file = new File(url.toURI());
        String line = IOHelper.getLine(file, 22);
		assertEquals(null, line);
	}

	@Test
	public void testReadEmptyFileAsBytes() throws Exception {
		File testFile = File.createTempFile("testFile", ".bin");
		assertArrayEquals(new byte[0], IOHelper.readFileAsBytes(testFile));
	}

	@Test(expected = IOException.class)
	public void testReadNonExistingFileAsBytes() throws Exception {
		File testFile = File.createTempFile("testFile", ".bin");
		FileSystemHelper.forceDelete(testFile);
		IOHelper.readFileAsBytes(testFile);
	}

	@Test
	public void testReadFileAsBytes() throws Exception {
		URL url = Thread.currentThread().getContextClassLoader()
				.getResource("binaryFileToRead.bin");
		File testFile = new File(url.toURI());
		int[] expectedBytesAsInt = new int[] { 0x01, 0x69, 0x00, 0x65, 0x31,
				0x0A, 0x20, 0x21, 0x25, 0x46, 0x54, 0x65, 0x45, 0x21, 0x68,
				0x94, 0x61, 0x56, 0xCA, 0xB5, 0x64, 0x61, 0x32, 0x1C, 0xBF };
		byte[] expectedBytes = new byte[expectedBytesAsInt.length];
		for (int i = 0; i < expectedBytesAsInt.length; ++i) {
			expectedBytes[i] = (byte) expectedBytesAsInt[i];
		}
		byte[] readBytes = IOHelper.readFileAsBytes(testFile);
		assertArrayEquals(expectedBytes, readBytes);
	}

	@Test
	public void testReadEmptyFileAsString() throws Exception {
		File testFile = File.createTempFile("testFile", ".txt");
		assertEquals("", IOHelper.readFileAsString(testFile));
	}

	@Test
	public void testReadFileAsString() throws Exception {
		URL url = Thread.currentThread().getContextClassLoader()
				.getResource("fileToRead.txt");
		File file = new File(url.toURI());
		String fileContent = IOHelper.readFileAsString(file);
        assertTrue(fileContent.endsWith("line16\n    line17   \n---line18---\nlast_line"));
	}

	@Test(expected = IOException.class)
	public void testReadNonExistingFileAsString() throws Exception {
		File testFile = File.createTempFile("testFile", ".txt");
		FileSystemHelper.forceDelete(testFile);
		IOHelper.readFileAsString(testFile);
	}

	@Test
	public void testReadFileAsStringWithEmptyLinesAtEnd() throws Exception {
		URL url = Thread.currentThread().getContextClassLoader()
				.getResource("fileToReadWithEmptyLinesAtEnd.txt");
		File file = new File(url.toURI());
		String fileContent = IOHelper.readFileAsString(file);
        assertTrue(fileContent.endsWith("line16\n    line17   \n---line18---\nlast_line\n\n\n"));
	}
}